/*
    $Id: dump.c,v 1.8 2001/11/22 15:18:33 sensei Exp $

    Copyright (C) 2000-2001  Marko Zivanovic
 
    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of the
    License, or (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
    02111-1307 USA
*/
     
#ifndef lint
static char vcid[] __attribute__ ((__unused__)) = "$Id: dump.c,v 1.8 2001/11/22 15:18:33 sensei Exp $";
#endif /* lint */

#include "dump.h"

int no_dumpers = -1;
dumper dumpers[MAX_DUMPERS];

int add_dumper (ClientData cd, Tcl_Interp* interp, int argc, char* argv[])
{
  int ret;
  int x;
  int df;

  if (argc != 3) {
/*      Tcl_WrongNumArgs (interp, argc, argv, "dumperID script"); */
    
    return TCL_ERROR;
  }
  
  df = 0;
  for (x = 0; x <= no_dumpers; x++) {
    if (dumpers[x].state == DSTATE_DELETED) {
      df = 1;
      break;
    }
  }

  if (df == 0) {
    if (no_dumpers == MAX_DUMPERS)
      return -1;
    
    no_dumpers++;
    x = no_dumpers;
  }
  
  dumpers[x].ID = (char*) malloc (strlen (argv[1]) + 1);
  if (dumpers[x].ID == NULL) {
    Tcl_AddErrorInfo (interp, "memory allocation error");
    return TCL_OK;
  }
  strcpy (dumpers[x].ID, argv[1]);
  
  dumpers[x].script = (char*) malloc (strlen (argv[2]) + 1);
  if (dumpers[x].script == NULL) {
    Tcl_AddErrorInfo (interp, "memory allocation error");
    return TCL_OK;
  }
  strcpy (dumpers[x].script, argv[2]);
  
  dumpers[x].state = DSTATE_RUNNING;
  
  /* 
     dumper.interp will be filled in by the dumper thread, after it is
     created 
  */
  
  ret = pthread_create (&dumpers[x].thread, NULL, start_dumper,
			dumpers[x].ID);
  if (ret != 0) {
    Tcl_AddErrorInfo (interp, "error creating thread");
    return TCL_ERROR;
  }

  ret = pthread_detach (dumpers[x].thread);
  if (ret != 0) {
    Tcl_AddErrorInfo (interp, "error detaching from thread");
    return TCL_ERROR;
  }

  return TCL_OK;
}

int pause_dumper (ClientData cd, Tcl_Interp* interp, int argc, char *argv[])
{
  dumper* tmp;

  if (argc != 2) {
    //    Tcl_WrongNumArgs (interp, argc, argv, "dumperID");
    return TCL_ERROR;
  }

  tmp = find_dumper (argv[1]);
  if (tmp == NULL) {
    Tcl_AddErrorInfo (interp, "dumper not found");
    return TCL_ERROR;
  }

  tmp->state = DSTATE_PAUSED;

  return TCL_OK;
}

int resume_dumper (ClientData cd, Tcl_Interp* interp, int argc, char* argv[])
{
  dumper* tmp;

  if (argc != 2) {
    //    Tcl_WrongNumArgs (interp, argc, argv, "dumperID");
    return TCL_ERROR;
  }

  tmp = find_dumper (argv[1]);
  if (tmp == NULL) {
    Tcl_AddErrorInfo (interp, "dumper not found");
    return TCL_ERROR;
  }

  tmp->state = DSTATE_RUNNING;

  return TCL_OK;
}

int delete_dumper (ClientData cd, Tcl_Interp* interp, int argc, char *argv[])
{
  dumper* tmp;

  if (argc != 2) {
    //    Tcl_WrongNumArgs (interp, argc, argv, "dumperID");
    return TCL_ERROR;
  }

  tmp = find_dumper (argv[1]);
  if (tmp == NULL) {
    Tcl_AddErrorInfo (interp, "dumper not found");
    return TCL_ERROR;
  }

  tmp->state = DSTATE_DELETE;

  return TCL_OK;
}

dumper* find_dumper (char* dumperID)
{
  int x;
  pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;

  pthread_mutex_lock (&mutex);
  for (x = 0; x <= no_dumpers; x++) {
    if (strcmp (dumpers[x].ID, dumperID) == 0)
      return &dumpers[x];
  }
  pthread_mutex_unlock (&mutex);

  return NULL;
}

/*  ************************************************************ */
/*  dumper thread main controling function                       */
/*                                                               */
/*  parameters:                                                  */
/*                                                               */
/*     param   -   parameter specified during thread creation    */
/*                 (currently unused)                            */
/*                                                               */
/*  return:                                                      */
/*     void*   -   return value; picked up with pthread_join     */
/*  ************************************************************ */
void *start_dumper (void* threadID)
{
  time_t last_dump, now;
  long pkt_count;
  Tcl_Interp* interp;
  dumper* myself;

  interp = Tcl_CreateInterp ();

  myself = find_dumper (threadID);
  if (myself == NULL) {
    logmsg (LOG_CRIT, "dumper thread not found: %s", (char*) threadID);
    exit (2);
  }

  myself->interp = interp;

  S_load_script (interp, myself->script);

  myself->priority = S_get_int_var (interp, "sniffer::dumper_priority");
  myself->dump_freq = S_get_int_var (interp, "sniffer::dumper_freq");
  myself->dump_size = S_get_int_var (interp, "sniffer::dumper_buffer_size");
  myself->dump_packets = S_get_int_var (interp, "sniffer::dumper_packets");

  S_call (interp, "sniffer::on_dumper_init");
  //  S_call (interp, "on_dumper_resume");

  while ((stop_dumper == 0) && (myself->state != DSTATE_DELETE)) {
    /* be extremely nice while idle */
    setpriority (PRIO_PROCESS, 0, 20);
    time (&last_dump);
    pkt_count = myself->buffered_pkts;
    while (1) {
      if (stop_dumper == 1)
	break;

      sleep (1);
      
      if (immediate_dump == 1) {
	immediate_dump = 0;
	break;
      }

      /* check if thread should pause execution */
      if (myself->state == DSTATE_PAUSED) {
	S_call (interp, "sniffer::on_dumper_pause");
	while (myself->state == DSTATE_PAUSED) {
	  /* do nothing */
	}
	S_call (interp, "sniffer::on_dumper_resume");
      }
      
      /* check if more than dump_freq seconds have elapsed since the
         last dump and if so, dump the buffer */
      time (&now);
      if (abs (difftime (now, last_dump)) >= myself->dump_freq) {
	break;
      }
      
      /* check if buffer has grown beyond dump_buffer_size limit; if
         it has, dump the buffer */
      if (myself->buffered_mem >= myself->dump_size * 1024) {
	break;
      }

      /* check if number of buffered packets has exceeded dump_packets
         count and if it has, dump buffer */
      if (myself->buffered_pkts - pkt_count >= myself->dump_packets) {
	break;
      }
    }
    
    setpriority (PRIO_PROCESS, 0, dumper_priority);
    dump_buffer (interp, (char*) threadID);
  }
  
  /* call dump_buffer once again just in case there are some packets
     left in the buffer */
  dump_buffer (interp, (char*) threadID);    

  S_call (interp, "sniffer::on_dumper_shutdown");

  free (myself->ID);
  free (myself->script);
  myself->state = DSTATE_DELETED;

  pthread_exit (NULL);
}

/*  ************************************************************ */
/*  dump current contents of buffer                              */
/*                                                               */
/*  no parameters                                                */
/*  return:                                                      */
/*     int   -   unused for now                                  */
/*  ************************************************************ */
int dump_buffer (Tcl_Interp* interp, char* threadID)
{
  eth_ip_frame* eth_ip;
  chain_buffer* cb_curr;
  dumper* dmp;

  dmp = find_dumper (threadID);

  cb_curr = cb_first;
  
  while (1) {
    if (cb_last == NULL)
      break; /* buffer is empty => exit */

    if ((stop_dumper == 0) && (cb_curr == cb_last))
      break; /* we're still running and the buffer is empty => exit */
    
    if (strcmp (cb_curr->dumperID, threadID) != 0)
      break; /* this is not our packet */

    eth_ip = cb_curr->eth_ip_frame;
    
    /* call on_packet_dump function */
    S_on_dump_packet (interp, eth_ip);

    /* if we should stop, dump the last remaining packet */
    if ((stop_dumper == 1) && (cb_first == cb_last)) {
      free (cb_curr->eth_ip_frame);
      free (cb_curr);
      cb_last = cb_first = cb_curr = NULL;
      break;
    }
    
    /*  remove the packet from the buffer */
    cb_first = cb_curr->next;
    dmp->buffered_mem -= cb_curr->eth_ip_frame->length;
    dmp->buffered_pkts--;
    free (cb_curr->eth_ip_frame);
    free (cb_curr);
    cb_curr = cb_first;
    
  } /* while cb_curr->next != cb_last */
  return 0;
}
