/*
    $Id: sniffer.c,v 1.8 2001/11/22 15:18:33 sensei Exp $

    Copyright (C) 2000-2001  Marko Zivanovic
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
     
#ifndef lint
static char vcid[] __attribute__ ((__unused__)) = "$Id: sniffer.c,v 1.8 2001/11/22 15:18:33 sensei Exp $";
#endif /* lint */

#include "sniffer.h"

chain_buffer *cb_first;
chain_buffer *cb_last;
int sock;

/*  ************************************************************ */
/*  sniffer thread main controling function. it handles reading  */
/*  data from socket and storing it into linked list buffer      */
/*                                                               */
/*  parameters:                                                  */
/*     param   -   parameter suplied by the pthread_create       */
/*                                                               */
/*  return:                                                      */
/*     void*   -   value received by pthread_join                */
/*  ************************************************************ */
void *start_sniffer (void *param)
{
#ifdef USE_PCAP
  pcap_t* pcd;
  char pc_err[1024];
#define PC_BUFLEN 4096
#else
  int sock;
  int len;
  struct timeval t;
  int ret;
  fd_set fds;
#endif
  Tcl_Interp* interp;
  char* res;
  eth_ip_frame *eth_ip;

  /* initialization */
#ifdef USE_PCAP
  pcd = pcap_open_live (device, PC_BUFLEN, 0, 0, pc_err);
  if (pcd == NULL) {
    logmsg (LOG_CRIT, "pcap error: %s", pc_err);
    exit (1);
  }
#else
  sock = init_socket (device);
#endif

  cb_first = cb_last = NULL;
  stop_sniffer = 0;

  interp = Tcl_CreateInterp ();

  /* FIXME: sniffer script should be read from global script file */
  S_load_script (interp, sniffer_file);

  S_call (interp, "on_sniffer_init");

  while (1) {
    setpriority (PRIO_PROCESS, 0, 20);
    while (1) {
      if (stop_sniffer == 1) {
#ifdef USE_PCAP
	pcap_close (pcd);
#endif
	logmsg (LOG_DEBUG, "sniffer exit stopped");
	pthread_exit (NULL);
      }
      
#ifdef USE_PCAP
      setpriority (PRIO_PROCESS, 0, sniffer_priority);
      eth_ip = read_eth_ip_frame_pcap (pcd);
      if (eth_ip != NULL)
	break;
#else /* use builtin mechanism */
      /*
	check to see if data is available on sock within some time
      */
      FD_ZERO (&fds);
      FD_SET (sock, &fds);
      t.tv_sec = 1; t.tv_usec = 0;
      
      ret = select (sock + 1, &fds, NULL, NULL, &t);
      if (ret == -1) continue;
      
      if (ret > 0) {
	setpriority (PRIO_PROCESS, 0, sniffer_priority);      
	eth_ip = read_eth_ip_frame (sock, len);
	break;
      }
#endif
    }
    
    /*
      call on_receive_packet function to see what to do with the
      packet received
    */
    res = S_on_receive_packet (interp, (eth_ip_frame*) eth_ip);

    if (strcmp (res, "skip") != 0) {
      add_packet_to_buffer (res, eth_ip);
    }

    total_pkts_sniffed++;

  } /* while endless loop */
  
  logmsg (LOG_DEBUG, "sniffer exit");
  return NULL;
}

/*  ************************************************************ */
/*  insert packet into the linked list buffer                    */
/*                                                               */
/*  parameters:                                                  */
/*     eth_ip_frame   -   packet to add to buffer                */
/*                                                               */
/*  no return                                                    */
/*  ************************************************************ */
int add_packet_to_buffer (char* dumperID, eth_ip_frame *eth_ip)
{
  chain_buffer *cb_curr;
  dumper* dmp;

  dmp = find_dumper (dumperID);

  cb_curr = (chain_buffer*) malloc (sizeof (chain_buffer));
  if (cb_curr == NULL) {
    logmsg (LOG_CRIT, "memory allocation error: %s", strerror (errno));
    exit (1);
  }
  
  cb_curr->eth_ip_frame = eth_ip;
  cb_curr->dumperID = (char*) malloc (strlen (dumperID) + 1);
  if (cb_curr->dumperID == NULL) {
    logmsg (LOG_CRIT, "memory allocation error: %s", strerror (errno));
    exit (2);
  }

  strcpy (cb_curr->dumperID, dumperID);

  if (cb_first == NULL)
    cb_first = cb_curr;
  else
    cb_last->next = cb_curr;
  
  cb_last = cb_curr;
  
  dmp->buffered_mem += eth_ip->length;
  total_mem_buffered += eth_ip->length;
  dmp->buffered_pkts++;
  total_pkts_buffered++;

  return 0;
}

#ifdef HAVE_PCAP_H
/*  ************************************************************ */
/*  read ethernet frame from socket using pcap library functions */
/*                                                               */
/*  parameters:                                                  */
/*     pcd    -   pcap device descriptor                         */
/*                                                               */
/*  return:                                                      */
/*     eth_ip_frame*   - Ethernet frame                          */
/*  ************************************************************ */
eth_ip_frame* read_eth_ip_frame_pcap (pcap_t* pcd)
{
  const u_char* pkt;
  struct pcap_pkthdr hdr;
  eth_ip_frame* eth_ip;

  pkt = pcap_next (pcd, &hdr);
  if (pkt == NULL) {
    return NULL;
  }

  eth_ip = (eth_ip_frame*) malloc (sizeof (eth_ip_frame));
  if (eth_ip == NULL) {
    logmsg (LOG_CRIT, "memory allocation error: %s", strerror (errno));
    exit (3);
  }

  /* convert pcap header into sniffer internal representation */
  eth_ip->length = hdr.caplen;
  eth_ip->time = hdr.ts.tv_sec;

  /* copy packet data into sniffer internal structures */
  memcpy (&eth_ip->eth, pkt, hdr.caplen);

  return eth_ip;
}
#endif

/*  ************************************************************ */
/*  read ethernet frame from socket                              */
/*                                                               */
/*  parameters:                                                  */
/*     sock   -   socket to read eth frames from                 */
/*     len    -   return parameter: size (in bytes) of the       */
/*                packet read                                    */
/*                                                               */
/*  return:                                                      */
/*     eth_ip_frame*   - Ethernet frame                          */
/*  ************************************************************ */
eth_ip_frame *read_eth_ip_frame (int sock, int len)
{
  eth_ip_frame *frame;
  char buffer[10240];
  int rlen;

  while (1) {
    len = read (sock, buffer, sizeof (buffer));
    if (len > 1) {
      rlen = len;
      /* check to see if we received proper eth frame */
      len = len - sizeof (struct ethhdr);
      if (len < 1) continue;
   
      /* time and size fields in the frame header */
      rlen += sizeof (struct tm) + sizeof (int);
      frame = (eth_ip_frame*) malloc (rlen);
      memcpy (&frame->eth, buffer, rlen - sizeof (struct tm) - sizeof (int));

      time (&frame->time);
      frame->length = rlen;

      return frame;
    }
  }
}

/*  ************************************************************ */
/*  create raw eth level socket and set device to promiscuous    */
/*  mode                                                         */
/*                                                               */
/*  parameters:                                                  */
/*     dev   -   device name                                     */
/*                                                               */
/*  return:                                                      */
/*     socket file descriptor                                    */
/*  ************************************************************ */
int init_socket (char *dev)
{
  sock = socket (PF_PACKET, SOCK_RAW, htons (ETH_P_ALL));
  if (sock < 0) {
    logmsg (LOG_CRIT, "error creating socket: %s", strerror (errno));
    exit (1);
  }

  return sock;
}

/*  ************************************************************ */
/*  switch device to promiscous mode                             */
/*                                                               */
/*  parameters:                                                  */
/*     sock   -   raw packet socket bound to interface           */
/*     dev    -   name of interface to switch to promiscous mode */
/*                                                               */
/*  no return value                                              */
/*  ************************************************************ */
void promisc_on (int sock, char *dev)
{
#ifndef USE_PCAP
  int res;
  struct ifreq ifr;
  
  /*    get interface flags */
  strcpy (ifr.ifr_name, dev);
  res = ioctl (sock, SIOCGIFFLAGS, &ifr);
  if (res < 0) {
    close (sock);
    logmsg (LOG_CRIT, "error in ioctl call: %s", strerror (errno));
    exit (1);
  }
  
  /*    switch interface to promiscuous mode */
  ifr.ifr_flags |= IFF_PROMISC;
  res = ioctl (sock, SIOCSIFFLAGS, &ifr);
  if (res < 0) logmsg (LOG_WARNING, "error switching interface to promiscuous mode: %s", strerror (errno));
#endif
}

/*  ************************************************************ */
/*  switch device from promiscous mode                           */
/*                                                               */
/*  parameters:                                                  */
/*     sock   -   raw packet socket bound to interface           */
/*     dev    -   name of interface to switch from promiscous    */
/*                mode                                           */
/*                                                               */
/*  no return value                                              */
/*  ************************************************************ */
void promisc_off (int sock, char *dev)
{
#ifndef USE_PCAP
  int res;
  struct ifreq ifr;
  
  /* get interface flags */
  strcpy (ifr.ifr_name, dev);
  res = ioctl (sock, SIOCGIFFLAGS, &ifr);
  if (res < 0) {
    close (sock);
    logmsg (LOG_CRIT, "error in ioctl call: %s", strerror (errno));
    exit (1);
  }
  
  /* switch interface from promiscuous mode */
  ifr.ifr_flags ^= IFF_PROMISC;
  res = ioctl (sock, SIOCSIFFLAGS, &ifr);
  if (res < 0) logmsg (LOG_WARNING, "error switching interface from promiscuous mode: %s", strerror (errno));
#endif
}
