/*
    $Id: vshare.c,v 1.1 2001/11/14 09:47:44 sensei Exp $

    Copyright (C) 2001  Wojciech Kocjan
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "vshare.h"

int vshare_tcl_cmd(ClientData cd,
    Tcl_Interp *interp,
    int objc,
    Tcl_Obj *const objv[])
{
    int cmd,rc;
    int ret; /* return code for mutex un/locking */

    if (objc<2)
    {
        Tcl_WrongNumArgs(interp,objc,objv,"command ?args?");
        return((int) TCL_ERROR);
    }
    if ((rc=Tcl_GetIndexFromObj(interp,objv[1],vs_cmds,"option",0,&cmd))!=TCL_OK)
    {
        return(rc);
    }

    /* 14 Nov 2001 sensei - added mutex locking */
    ret = pthread_mutex_lock (&S_mutex);
    if (ret != 0) {
      logmsg (LOG_CRIT, "mutex error: %s", strerror (errno));
      exit (2);
    }

    switch (cmd) {
        case CMD_UNSET:
            {
	      struct vs_value *vv; /* unused: Tcl_Obj *ro; */
                if (objc!=4)
                {
                    Tcl_WrongNumArgs(interp,2,objv,"array variable");
                    rc=TCL_ERROR;
                }
                else
                {
                    if ((vv=vs_find_entry(Tcl_GetString(objv[2]),Tcl_GetString(objv[3]),0))!=NULL)
                    {
                        vs_del_val(vv);
                    }
                    
                }
                break;
            }
        case CMD_EVAL:
            {
                struct vs_value *vv; int l; char *vc;
                if (objc!=7)
                {
                    Tcl_WrongNumArgs(interp,2,objv,"array variable tmpvar defval code");
                    rc=TCL_ERROR;
                }
                else
                {
                    if ((vv=vs_find_entry(Tcl_GetString(objv[2]),Tcl_GetString(objv[3]),1))!=NULL)
                    {
                        if (vv->val)
                        {
                            Tcl_ObjSetVar2(interp,objv[4],NULL,Tcl_NewStringObj(vv->val,vv->len),0);
                        }
                        else
                        {
                            Tcl_ObjSetVar2(interp,objv[4],NULL,objv[5],0);
                        }
                        if (Tcl_EvalObjEx(interp,objv[6],0)==TCL_OK)
                        {
                            vc=Tcl_GetStringFromObj(Tcl_ObjGetVar2(interp,objv[4],NULL,0),&l);
                            vs_del_val(vv);
                            vs_set_val(vv,vc,l);
                        }
                        Tcl_UnsetVar2(interp,Tcl_GetString(objv[4]),NULL,0);
                    }
                }
                break;
            }
        case CMD_SET:
            {
                struct vs_value *vv; int l; char *vc;
                if (objc!=5)
                {
                    Tcl_WrongNumArgs(interp,2,objv,"array variable newval");
                    rc=TCL_ERROR;
                }
                else
                {
                    if ((vv=vs_find_entry(Tcl_GetString(objv[2]),Tcl_GetString(objv[3]),1))!=NULL)
                    {
                        vc=Tcl_GetStringFromObj(objv[4],&l);
                        vs_del_val(vv);
                        vs_set_val(vv,vc,l);
                    }
                }
                break;
            }
        case CMD_GET:
            {
                struct vs_value *vv; Tcl_Obj *ro;
                if (objc!=4)
                {
                    Tcl_WrongNumArgs(interp,2,objv,"array variable");
                    rc=TCL_ERROR;
                }
                else
                {
                    if ((vv=vs_find_entry(Tcl_GetString(objv[2]),Tcl_GetString(objv[3]),0))!=NULL)
                    {
                        if (vv->val)
                        {
			  if ((ro=Tcl_NewStringObj(vv->val,vv->len)))
                            {
                                Tcl_SetObjResult(interp,ro);
                            }
                        }
                        else
                        {
                            Tcl_SetResult(interp,"Variable not found",TCL_STATIC);
                            rc=TCL_ERROR;
                        }
                    }
                    
                    else
                    {
                        Tcl_SetResult(interp,"Variable not found",TCL_STATIC);
                        rc=TCL_ERROR;
                    }
                }
                break;
            }
        default:
            Tcl_SetResult(interp,"Command not implemented",TCL_STATIC);
            rc=TCL_ERROR;
    }

    /* 14 Nov 2001 sensei */
    ret = pthread_mutex_unlock (&S_mutex);
    if (ret != 0) {
      logmsg (LOG_CRIT, "mutex error: %s", strerror (errno));
      exit (2);
    }

    return(rc);
}

void vs_init_val(struct vs_value *v)
{
    v->val=NULL;
    v->len=0;
}
void vs_set_val(struct vs_value *v, char *nv, int l)
{
    if (v->val!=NULL)
    {
        free(v->val);
        v->val=NULL;
    }
    if (l<0)
        l=strlen(nv);
    v->val=malloc(l+1);
    memcpy(v->val,nv,l+1);
    v->len=l;
}
void vs_del_val(struct vs_value *v)
{
    if (v->val!=NULL)
    {
        free(v->val);
        v->val=NULL;
    }
    v->len=0;
}
struct vs_value *vs_find_entry(char *ar, char *val, int mknew)
{
    Tcl_HashEntry *he;
    int new;
    char buf[1024];
    struct vs_value *rc=NULL;
    
    sprintf(buf,"%s:%s",ar,val);
    if (mknew)
    {
        if ((he=Tcl_CreateHashEntry(&vs_hash,buf,&new))!=NULL)
        {
            if (new)
            {
                rc=malloc(sizeof(struct vs_value));
                rc->len=0;
                rc->val=NULL;
                Tcl_SetHashValue(he,rc);
            }
            else
            {
                rc=Tcl_GetHashValue(he);
            }
        }
    }
    else
    {
      if ((he=Tcl_FindHashEntry(&vs_hash,buf)))
        {
            rc=Tcl_GetHashValue(he);
        }
    }
    
    return(rc);
}

void vshare_globalinit(void)
{
    Tcl_InitHashTable(&vs_hash,TCL_STRING_KEYS);
}

int Vshare_Init(Tcl_Interp *ti)
{
    vshare_globalinit();
    Tcl_CreateObjCommand(ti,"vshare",vshare_tcl_cmd,NULL,NULL);

    /* 14 Nov 2001 sensei */
    return 0;
}
